"""
Entrypoint for running an instance of the cli. Used by setuptools when installing an entrypoint.

Functions:
    init - Initialise the CLI
"""
from .cli import CLI


def init():
    """Initalise the CLI"""
    CLI()



"""
Constant values used in the project

Variables:
    VERSION           - The version of the program as a string
    BASE_API_URL      - The root URL of the API
    USER_AGENT        - The default user-agent to send with every request
    METADATA_FILENAME - The filename for metadata files saved in jars
"""
# Static values
VERSION = "1.3.0"
BASE_API_URL = "https://api.spiget.org/v2"
USER_AGENT = f"Spud/{VERSION}"
METADATA_FILENAME = ".spud_meta.json"



"""
Module for CLI interfaces to the API backend

classes:
    CLI - Represents a CLI interface

"""
from __future__ import annotations

import sys
from argparse import ArgumentParser, Namespace
import os
from pathlib import Path
from typing import Collection, Union

from . import api
from .utils import Utils, Color
from .type import StatusDict, Plugin, Metadata, Update
from .settings import VERSION


class CLI:
    """Represents the spud CLI. Handles program arguments and allows the user to interact with the API"""

    def __init__(self, api_class=api.SpigetAPI) -> None:
        """Initialise the cli application"""
        self.api = api_class()

        self.args: Namespace = self.parse_args()

        os.chdir(self.args.directory)

        if self.args.action == "install":
            self.install(self.args.plugins)
        elif self.args.action == "update":
            self.update(self.args.plugins)
        else:
            Utils.format_text(f"Action {self.args.action} does not exist", Color.ERROR)

    def install(self, plugins: Collection[str]) -> None:
        """
        Download the jars for a list of plugins and save them as files.

        :param plugins: A list of plugin names or jar filenames
        """
        for plugin_name in plugins:
            plugin_name = Utils.get_plugin_name_from_jar(plugin_name)

            plugin_list = self.api.search_plugins(plugin_name)

            if not plugin_list:
                Utils.format_text(
                    f"No plugin with name {plugin_name} found.", Color.ERROR
                )
                continue

            Utils.format_text(f"Query: {plugin_name}", Color.SUCCESS)

            if self.args.noninteractive:
                plugin = plugin_list[0]
            else:
                plugin = self.get_plugin_choice(plugin_list)

            if not plugin:
                Utils.format_text("Skipping Install!", Color.ERROR)
                continue

            Utils.format_text(f"Installing {plugin['name']}", Color.STATUS)

            result: StatusDict = self.api.download_plugin(plugin)

            if result["status"]:
                Utils.format_text(
                    f"{plugin['name']} was installed successfully",
                    Color.SUCCESS,
                )
            else:
                Utils.format_text(result["message"], Color.WARNING)

    def update(self, plugins: Collection[str]) -> None:
        """
        Check if plugins need updates, and if they do update them.

        :param plugins: An list of plugin names or jar filenames
        """
        if not plugins:
            file_list = os.listdir()
            plugins = [i for i in file_list if i.endswith(".jar")]
            Utils.format_text(
                f"Detected {len(plugins)} plugins in {os.getcwd()}", Color.STATUS
            )

        update_count = 0
        for plugin_name in plugins:
            filename = plugin_name
            if ".jar" not in filename:
                filename = Utils.create_jar_name(plugin_name)

            metadata: Union[Metadata, None] = Utils.load_metadata_file(filename)

            if not metadata:
                Utils.format_text(
                    f"Couldn't load metadata for {filename}. Try reinstalling with spud first",
                    Color.WARNING,
                )
                continue

            plugin: Union[Plugin, None] = self.api.get_plugin_info_if_update(metadata)
            if plugin is None:
                Utils.format_text(
                    f"Plugin {metadata['search_name']} already up to date", Color.STATUS
                )
                continue

            # If running in interactive mode
            if not self.args.noninteractive:
                update: Update = self.api.get_latest_update_info(plugin)
                changelog = update["description"]
                Utils.separator()
                Utils.format_text(f"Changelog for {plugin['name']}:", Color.STATUS)
                Utils.format_text(changelog, Color.STATUS)
                Utils.separator()

                # If user doesn't want to update
                if not Utils.prompt_bool(f"Would you like to update {plugin['name']}?"):
                    Utils.format_text(f"Not updating {plugin['name']}", Color.WARNING)
                    continue

            # By this point we can ensure that:
            #  1. The plugin has an update available
            #  2. The user wants to update

            # Download the latest version of the plugin
            result: StatusDict = self.api.download_plugin(plugin, filename)

            color = Color.WARNING
            # If the download succeeded
            if result["status"]:
                update_count += 1
                color = Color.SUCCESS

            if message := result["message"]:
                Utils.format_text(message, color)

        Utils.separator()
        Utils.format_text(
            f"{update_count} updated, {len(plugins) - update_count} left unchanged",
            Color.STATUS,
        )

    @staticmethod
    def parse_args(argv=None) -> Namespace:
        """
        :arg argv: A list of arguments to parse, defaults to sys.argv
        :return: a Namespace instance for the program's arguments
        """
        if argv is None:
            argv = sys.argv

        # Remove program name from argv
        argv = argv[1:]

        parser = ArgumentParser(
            description="Spud: The plugin manager for your Spigot Minecraft server",
            epilog="Licensed under GPLv3 (https://www.gnu.org/licenses/gpl-3.0.en.html). "
            "Source available at https://github.com/exciteabletom/spud",
        )
        parser.add_argument("action", help="install or update", type=str)
        parser.add_argument(
            "plugins",
            metavar="name",
            help="the names (or filenames) of plugins",
            type=str,
            nargs="*",
        )
        parser.add_argument(
            "-n",
            dest="noninteractive",
            help="run without asking for input",
            action="store_true",
            default=False,
        )
        parser.add_argument(
            "-d",
            "--directory",
            dest="directory",
            help="path to plugins directory, defaults to the working directory",
            type=Path,
            default=".",
        )
        parser.add_argument(
            "-v", "--version", action="version", version=f"%(prog)s {VERSION}"
        )
        return parser.parse_args(argv)

    @staticmethod
    def get_plugin_choice(plugin_list: list[Plugin]) -> Union[Plugin, None]:
        """
        Gets a user's choice out of a list of plugins

        :returns: The Plugin dict the user chose, or None if they didn't choose one
        """
        Utils.separator()
        for index, plugin in enumerate(plugin_list):
            Utils.format_text(
                f"{index} | {plugin['name']} by {plugin['author']['name']} | {plugin['tag']}",
                Color.STATUS,
            )

        Utils.separator()
        while True:
            try:
                chosen_id: int = int(
                    Utils.prompt("Select a plugin ID (Ctrl-D to skip)")
                )
            except ValueError:
                continue
            except EOFError:
                return None

            if 0 <= chosen_id < len(plugin_list):
                return plugin_list[chosen_id]

            continue



"""
Custom types used in the program.

TypedDicts:
    StatusDict
    Author
    PluginVersion
    Plugin
    Metadata
    Update
"""
from typing import TypedDict


class StatusDict(TypedDict):
    """Allows you to return a status value and an error/success message from a function"""

    status: bool
    message: str


class Author(TypedDict, total=False):
    """Represents a resource Author"""

    id: int
    name: str


class PluginVersion(TypedDict):
    """Represents a specific plugin version"""

    id: int
    uuid: str


class Plugin(TypedDict):
    """Represents a plugin"""

    file: dict
    name: str
    tag: str
    version: PluginVersion
    id: int
    str: int
    author: Author


class Metadata(TypedDict):
    """Represents the metadata values saved into jar files"""

    search_name: str
    plugin_id: int
    plugin_version_id: int


class Update(TypedDict):
    """Represents a specific update of a plugin received"""

    id: int
    title: str
    description: str
    date: int
    likes: int



"""
Classes for interacting with APIs.

classes:
    SpigetAPI - Helps interacting with the Spiget API
"""
from __future__ import annotations

import base64
import itertools
import operator
from typing import Union

from bs4 import BeautifulSoup

import requests
from requests import HTTPError

from . import settings
from .type import Plugin, StatusDict, Author, Metadata, Update
from .utils import Utils


class SpigetAPI:
    """
    A class to represent a connection with the Spiget API
    """

    def __init__(
        self,
        base_api_url: str = settings.BASE_API_URL,
        user_agent: str = settings.USER_AGENT,
    ) -> None:
        """
        Initialise an instance of the Spiget API

        :param base_api_url: The root API http URL, default: settings.BASE_API_URL
        :param user_agent: The user-agent header to send with requests, default: settings.USER_AGENT
        """
        self.base_api_url = base_api_url

        self.headers: dict = {
            "user-agent": user_agent,
        }

    def build_api_url(self, endpoint: str) -> str:
        """Append an endpoint to the base API url and return it"""
        return f"{self.base_api_url}{endpoint}"

    def call_api(self, endpoint: str, params=None) -> requests.Response:
        """
        Call an API endpoint

        :param endpoint: The endpoint to call
        :param params: Request body as a dict (optional)
        :returns: A Response object
        """
        if params is None:
            params = {}

        response = requests.get(
            self.build_api_url(endpoint),
            params=params,
            headers=self.headers,
        )
        if str(response.status_code).startswith("5"):
            raise HTTPError

        return response

    def get_plugin_by_id(self, plugin_id: int) -> Plugin:
        """
        Get a Plugin using an ID.

        :returns: A dict of type Plugin
        """
        response = self.call_api(f"/resources/{plugin_id}")
        return response.json()

    def search_plugins(self, query: str) -> list[Plugin]:
        """
        Search for plugins using a query

        :returns: A list of Plugin dicts, sorted by relevance to the query
        """
        names = [query]
        split_name: str = Utils.split_title_case(query)
        if split_name:
            names.append(split_name)

        plugin_list: list[Plugin] = []

        for name in names:
            response = self.call_api(
                f"/search/resources/{name}",
                {
                    "field": "name",
                    "sort": "-downloads",
                    "size": 5,
                    "fields": "file,name,tag,version,downloads,id,author",
                },
            )
            if response.status_code == 200:
                plugin: Plugin = response.json()[0]
                plugin_list.append(plugin)

        if len(plugin_list) == 0:
            return []

        # Sort the list by highest downloads, then IDs
        plugin_list.sort(key=operator.itemgetter("downloads", "id"), reverse=True)

        # Remove duplicate ids from list
        plugin_list = [id_field[0] for id_field in itertools.groupby(plugin_list)]

        sorted_list: list[Plugin] = []
        for plugin in plugin_list:
            # Exact match goes to first index
            if query.upper() == plugin["name"].upper():
                sorted_list.insert(0, plugin)

            # Fuzzy match goes to second index
            elif query.upper() in plugin["name"].upper():
                sorted_list.insert(1, plugin)

            # Everything else just gets appended to the end
            else:
                sorted_list.append(plugin)

        truncated_list: list[Plugin] = sorted_list[:10]

        for plugin in truncated_list:
            Utils.sanitise_api_plugin(plugin)
            # Author's name ==
            plugin["author"]["name"] = self.get_author(
                # Author's ID
                plugin["author"]["id"]
            )["name"]

        return truncated_list

    def download_plugin(self, plugin: Plugin, filename: str = "") -> StatusDict:
        """
        Download a plugin

        :param plugin: Plugin dict
        :param filename: Force a filename for the plugin instead of inferring it
        :return: StatusDict
        """
        response = self.call_api(
            f"/resources/{plugin['id']}/download",
        )
        if response.status_code != 200:
            return {
                "status": False,
                "message": "Could not download resource due to an unknown error. "
                "This can sometimes happen with external resources.",
            }

        if not filename:
            plugin_jar_name = Utils.create_jar_name(plugin["name"])
        else:
            plugin_jar_name = filename

        with open(plugin_jar_name, "wb") as file:
            file.write(response.content)

        Utils.inject_metadata_file(plugin, plugin_jar_name)

        return {"status": True, "message": f"Downloaded {plugin_jar_name}"}

    def get_plugin_info_if_update(self, metadata: Metadata) -> Union[Plugin, None]:
        """
        Get a Plugin dict for a plugin if it has been updated.

        :param metadata: A Metadata dict
        :returns: A Plugin dict if there was an update, otherwise None
        """
        plugin_id: int = metadata["plugin_id"]
        plugin = self.get_plugin_by_id(plugin_id)

        local_version: int = metadata["plugin_version_id"]
        latest_version: int = plugin["version"]["id"]

        if local_version >= latest_version:
            return None

        return plugin

    def get_author(self, author_id: int) -> Author:
        """
        Gets an Author dict from an ID

        :param author_id: The ID of an author
        :return: A dict representing the author
        """
        return self.call_api(f"/authors/{author_id}").json()

    def get_latest_update_info(self, plugin: Plugin) -> Update:
        """
        Get information about the latest update of a plugin

        :param plugin: A Plugin dict
        :return: An Update dict
        """
        response = self.call_api(f"/resources/{plugin['id']}/updates/latest")
        response.raise_for_status()

        update: Update = response.json()
        # Decode base64
        update["description"] = bytes.decode((base64.b64decode(update["description"])))
        # Convert html to plaintext
        update["description"] = " ".join(
            BeautifulSoup(update["description"], "html.parser").stripped_strings
        )
        # Remove duplicate newlines and spaces
        update["description"] = update["description"].replace("  ", " ")
        update["description"] = update["description"].replace("\r\n", "\n")
        update["description"] = update["description"].replace("\n\n", "\n")

        if len(update["description"]) > 1500:
            update["description"] = update["description"][:1500] + "..."
        return update



"""
setup.py file for installing Spud with pip
"""
import setuptools

from spud import settings

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

requirements = []
with open("requirements.txt", "r", encoding="utf-8") as f:
    for line in f:
        requirements.append(line.strip())

setuptools.setup(
    name="spud-mc",
    version=settings.VERSION,
    author="Tommy Dougiamas",
    author_email="tom@digitalnook.net",
    description="A Minecraft Spigot plugin manager that adheres to the Unix philosophy and Python best practices",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/exciteabletom/spud",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: System Administrators",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Topic :: System :: Software Distribution",
        "Typing :: Typed",
    ],
    entry_points={
        "console_scripts": ["spud = spud.__main__:init"],
    },
    package_dir={"spud": "spud"},
    packages=["spud"],
    install_requires=requirements,
    python_requires=">=3.8",
)
